#define _GNU_SOURCE
#include "recv_packet.h"

#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <ifaddrs.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/errqueue.h>
#include <linux/ethtool.h>
#include <linux/net_tstamp.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <poll.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "common.h"
#include "tracer.h"

#ifdef WITH_XDP
#include <bpf/bpf.h>
#include <bpf/libbpf.h>
#include <bpf/xsk.h>
#include <linux/if_ether.h>
#include <linux/if_link.h>
#include <linux/if_xdp.h>
#include <linux/ip.h>
#include <linux/udp.h>
#endif

static char rx_buffer[MAX_BUFFER_SIZE];
static int sock_fd;

static int so_timestamping_flags =
    SOF_TIMESTAMPING_RX_SOFTWARE | SOF_TIMESTAMPING_SOFTWARE;

static ingress_param_t *params;
static ingress_stat_t *stats;
static uint64_t *kernel_latency_hist;
static int use_histogram;

static uint64_t post_kernel_timestamp;

static char ts_tracemark_buf[64];

// Sets the interface
static int set_if(void) {
  struct ifreq ifreq;

  memset(&ifreq, 0, sizeof(ifreq));
  strncpy(ifreq.ifr_name, params->network_if, sizeof(ifreq.ifr_name) - 1);

  if (ioctl(sock_fd, SIOCGIFINDEX, &ifreq))
    error(EXIT_FAILURE, errno, "ioctl SIOCGIFINDEX failed\n");

  return ifreq.ifr_ifindex;
}

void init_udp_recv(ingress_param_t *_params, ingress_stat_t *_stats,
                   int _use_histogram, uint64_t *_kernel_latency_hist) {
  int getaddrinfo_err;
  int set_if_err;
  struct addrinfo hints, *servinfo, *servinfo_it;

  params = _params;
  stats = _stats;
  use_histogram = _use_histogram;
  kernel_latency_hist = _kernel_latency_hist;

  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_flags = AI_PASSIVE;

  getaddrinfo_err = getaddrinfo(NULL, SERVER_PORT, &hints, &servinfo);
  if (getaddrinfo_err != 0) {
    fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(getaddrinfo_err));
    exit(EXIT_FAILURE);
  }

  for (servinfo_it = servinfo; servinfo_it;
       servinfo_it = servinfo_it->ai_next) {
    sock_fd = socket(servinfo->ai_family, servinfo->ai_socktype,
                     servinfo->ai_protocol);

    if (bind(sock_fd, servinfo_it->ai_addr, servinfo_it->ai_addrlen) == -1) {
      close(sock_fd);
      continue;
    }
    break;
  }

  freeaddrinfo(servinfo);

  if (sock_fd == -1)
    error(EXIT_FAILURE, errno, "Couldn't create receive socket");

  set_if_err = set_if();
  if (set_if_err < 0) error(EXIT_FAILURE, errno, "Couldn't set interface\n");

  if (setsockopt(sock_fd, SOL_SOCKET, SO_BINDTODEVICE, params->network_if,
                 strlen(params->network_if)))
    error(EXIT_FAILURE, errno, "setsockopt SO_BINDTODEVICE failed\n");

  if (params->use_timestamps) {
    if (setsockopt(sock_fd, SOL_SOCKET, SO_TIMESTAMPING, &so_timestamping_flags,
                   sizeof(so_timestamping_flags)))
      error(EXIT_FAILURE, errno, "setsockopt SO_TIMESTAMPING failed\n");
  }
}

/*
 * Receive UDP packet
 */
void recv_udp_packet() {
  struct cmsghdr *cmsg;
  struct msghdr msg;  // Message hardware, sent to the socket
  struct iovec iov;   // The iovec structures stores the RX buffer
  struct sockaddr_in sin;
  struct {
    struct cmsghdr cm;
    char control[512];
  } control;

  int recvmsgerr;

  struct timespec ts;

  iov.iov_base = &rx_buffer;
  iov.iov_len = MAX_BUFFER_SIZE - 1;

  memset(&msg, 0, sizeof(msg));
  msg.msg_name = &sin;
  msg.msg_namelen = sizeof(sin);
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = &control;
  msg.msg_controllen = sizeof(control);

  recvmsgerr = recvmsg(sock_fd, &msg, 0);
  if (recvmsgerr < 0)
    error(EXIT_FAILURE, errno, "recvmsg failed, ret value: %d\n", recvmsgerr);

  if (params->use_timestamps) {
    clock_gettime(CLOCK_REALTIME, &ts);
    post_kernel_timestamp = ts_to_uint(ts);
  }

  if (params->use_timestamps) {
    for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
      if (cmsg->cmsg_level == SOL_SOCKET &&
          cmsg->cmsg_type == SO_TIMESTAMPING) {
        struct timespec *stamp = (struct timespec *)CMSG_DATA(cmsg);

        if (params->enable_ts_tracemark) {
          sprintf(ts_tracemark_buf, "%" PRIu64, ts_to_uint(*stamp));
          tracemark(ts_tracemark_buf);
        } else {
          uint64_t kernel_latency = post_kernel_timestamp - ts_to_uint(*stamp);
          kernel_latency /= 1000u;

          stats->min_kernel_latency =
              _min_(kernel_latency, stats->min_kernel_latency);
          stats->max_kernel_latency =
              _max_(kernel_latency, stats->max_kernel_latency);
          stats->avg_kernel_latency =
              (stats->max_kernel_latency * (stats->packets_received) +
               kernel_latency) /
              (stats->packets_received + 1);

          if (use_histogram) {
            if (kernel_latency > MAX_KERNEL_LATENCY)
              stats->high_kernel_latency++;
            else
              kernel_latency_hist[kernel_latency]++;
          }
        }
      }
    }
  }

  for (int i = 0; i < MAX_BUFFER_SIZE; i++) stats->data[i] = rx_buffer[i];
}

#ifdef WITH_XDP
static int xdp_flags = XDP_FLAGS_DRV_MODE;
static struct pollfd fds[1] = {0};
static unsigned int ifindex;
static struct xdpsock xdp_socket;

static void open_xdp_socket(char *network_if) {
  struct xsk_socket_config xsk_cfg;
  uint32_t idx;
  int ret, i;

  /* Create XDP socket */
  xsk_cfg.rx_size = XSK_RING_CONS__DEFAULT_NUM_DESCS;
  xsk_cfg.tx_size = XSK_RING_PROD__DEFAULT_NUM_DESCS;
  xsk_cfg.libbpf_flags = 0;
  xsk_cfg.xdp_flags = xdp_flags;
  xsk_cfg.bind_flags = 0;

  ret = xsk_socket__create(&xdp_socket.xsk, network_if, 0, xdp_socket.umem.umem,
                           &xdp_socket.rx, &xdp_socket.tx, &xsk_cfg);
  if (ret) err("xsk_socket__create() failed");

  /* Add some buffers */
  ret = xsk_ring_prod__reserve(&xdp_socket.umem.fq,
                               XSK_RING_PROD__DEFAULT_NUM_DESCS, &idx);

  if (ret != XSK_RING_PROD__DEFAULT_NUM_DESCS)
    err("xsk_ring_prod__reserve() failed");

  for (i = 0; i < XSK_RING_PROD__DEFAULT_NUM_DESCS; i++)
    *xsk_ring_prod__fill_addr(&xdp_socket.umem.fq, idx++) = i * FRAME_SIZE;

  xsk_ring_prod__submit(&xdp_socket.umem.fq, XSK_RING_PROD__DEFAULT_NUM_DESCS);
}

/*
 * Init XDP socket
 */
void init_xdp_recv(ingress_param_t *_params, ingress_stat_t *_stats) {
  int ret, prog_fd, xsks_map = 0;
  struct bpf_prog_load_attr prog_load_attr = {
      .prog_type = BPF_PROG_TYPE_XDP,
      .file = "/home/oli/tsn-measures/packet-exchange/build/xdp_kern.o",
  };
  struct xsk_umem_config cfg = {
      .fill_size = XSK_RING_PROD__DEFAULT_NUM_DESCS,
      .comp_size = XSK_RING_CONS__DEFAULT_NUM_DESCS,
      .frame_size = FRAME_SIZE,
      .frame_headroom = XSK_UMEM__DEFAULT_FRAME_HEADROOM,
      .flags = 0,
  };
  struct bpf_object *obj;
  struct bpf_map *map;
  void *buffer = NULL;

  params = _params;
  stats = _stats;

  ret = bpf_prog_load_xattr(&prog_load_attr, &obj, &prog_fd);
  if (ret || prog_fd < 0) err("bpf_prog_load_xattr() failed");

  map = bpf_object__find_map_by_name(obj, "xsks_map");
  xsks_map = bpf_map__fd(map);
  if (xsks_map < 0) err("No xsks_map found!");

  ifindex = if_nametoindex(params->network_if);
  if (!ifindex) err_errno("if_nametoindex() failed");

  /* Use XDP _only_ in conjuction with driver assisted mode */
  ret = bpf_set_link_xdp_fd(ifindex, prog_fd, xdp_flags);
  if (ret) err("bpf_set_link_xdp_fd() failed");

  /* Allocate user space memory for xdp frames */
  ret =
      posix_memalign(&buffer, sysconf(_SC_PAGE_SIZE), NUM_FRAMES * FRAME_SIZE);
  if (ret) err_errno("posix_memalign() failed");

  ret = xsk_umem__create(&xdp_socket.umem.umem, buffer, NUM_FRAMES * FRAME_SIZE,
                         &xdp_socket.umem.fq, &xdp_socket.umem.cq, &cfg);
  if (ret) err("xsk_umem__create() failed");
  xdp_socket.umem.buffer = buffer;

  /* Open and bind socket */
  open_xdp_socket(params->network_if);
}

void setup_poll_fd(void) {
  fds[0].fd = xsk_socket__fd(xdp_socket.xsk);
  fds[0].events = POLLIN;
}

static int received;
static uint32_t idx_rx = 0, idx;

static void parse_raw_packet(uint64_t addr, size_t len) {
  char *packet;
  struct ethhdr *eth;
  struct iphdr *ip;
  struct udphdr *udp;
  size_t min_len = sizeof(*eth) + sizeof(*ip) + sizeof(*udp);

  if (len <= min_len) {
    stats->xdp_data = NULL;
    return;
  }

  packet = xsk_umem__get_data(xdp_socket.umem.buffer, addr);

  eth = (struct ethhdr *)packet;
  ip = (struct iphdr *)(packet + sizeof(*eth));
  udp = (struct udphdr *)(packet + sizeof(*eth) + sizeof(*ip));

  stats->xdp_data = packet + sizeof(*eth) + sizeof(*ip) + sizeof(*udp);
}

/*
 * Receive XDP socket
 */
int recv_xdp_packet(struct timespec next) {
  int ret;
  uint64_t addr;
  uint32_t len;
  struct timespec next_pre, current;
  int k = 0;

  ret = poll(fds, 1, -1);

  if (ret != 1)
    error(EXIT_FAILURE, errno, "poll failed");

  received = xsk_ring_cons__peek(&xdp_socket.rx, 1, &idx_rx);

  if (received != 1)
    error(EXIT_FAILURE, errno, "Received %d packets", received);

  /* Get the packet */
  addr = xsk_ring_cons__rx_desc(&xdp_socket.rx, idx_rx)->addr;
  len = xsk_ring_cons__rx_desc(&xdp_socket.rx, idx_rx)->len;

  /* Parse it */
  parse_raw_packet(xsk_umem__add_offset_to_addr(addr), len);

  return 0;
}
void recv_xdp_cleanup(void) {
  uint64_t addr;
  int ret;

  /* Cleanup */
  xsk_ring_cons__release(&xdp_socket.rx, received);

  /* Add that particular buffer back to the fill queue */
  if (xsk_prod_nb_free(&xdp_socket.umem.fq, received)) {
    ret = xsk_ring_prod__reserve(&xdp_socket.umem.fq, received, &idx);

    if (ret != received) err("xsk_ring_prod__reserve() failed");

    *xsk_ring_prod__fill_addr(&xdp_socket.umem.fq, idx) =
        xsk_umem__extract_addr(addr);

    xsk_ring_prod__submit(&xdp_socket.umem.fq, received);
  }
}

void close_xdp_socket(void) {
  xsk_socket__delete(xdp_socket.xsk);
  xsk_umem__delete(xdp_socket.umem.umem);
  bpf_set_link_xdp_fd(ifindex, -1, xdp_flags);
}

#else
void init_xdp_recv(ingress_param_t *_params, ingress_stat_t *_stats) {
  (void)_params;
  (void)_stats;
}
void setup_poll_fd(void) {}
void close_xdp_socket(void) {}
int recv_xdp_packet(struct timespec next) {
  (void)next;
  return 0;
}
void recv_xdp_cleanup(void) {}
#endif

#ifdef DEBUG

/*
 * Code from scheduled_tx_tools
 */
static int process_socket_error_queue(void) {
  uint8_t msg_control[CMSG_SPACE(sizeof(struct sock_extended_err))];
  unsigned char err_buffer[256];
  struct sock_extended_err *serr;
  struct cmsghdr *cmsg;
  __u64 tstamp = 0;

  struct iovec iov = {.iov_base = err_buffer, .iov_len = sizeof(err_buffer)};
  struct msghdr msg = {.msg_iov = &iov,
                       .msg_iovlen = 1,
                       .msg_control = msg_control,
                       .msg_controllen = sizeof(msg_control)};

  if (recvmsg(sock_fd, &msg, MSG_ERRQUEUE) == -1) {
    fprintf(stderr, "recvmsg failed");
    return -1;
  }

  cmsg = CMSG_FIRSTHDR(&msg);
  while (cmsg != NULL) {
    serr = (void *)CMSG_DATA(cmsg);
    if (serr->ee_origin == SO_EE_ORIGIN_TXTIME) {
      tstamp = ((__u64)serr->ee_data << 32) + serr->ee_info;

      switch (serr->ee_code) {
        case SO_EE_CODE_TXTIME_INVALID_PARAM:
          fprintf(stderr,
                  "packet with tstamp %llu dropped "
                  "due to invalid params\n",
                  tstamp);
          return 0;
        case SO_EE_CODE_TXTIME_MISSED:
          fprintf(stderr,
                  "packet with tstamp %llu dropped "
                  "due to missed deadline\n",
                  tstamp);
          return 0;
        default:
          return -1;
      }
    }

    cmsg = CMSG_NXTHDR(&msg, cmsg);
  }

  return 0;
}
#endif
