/*
 * Real time packet receiving server
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <inttypes.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <unistd.h>

#include "common.h"
#include "gpio.h"
#include "recv_packet.h"
#include "send_packet.h"
#include "tracer.h"

#define ERROR_MARGIN_NS 400000

// Structs

enum TSNTask { RECV_PACKET_TASK, RTT_TASK, XDP_TASK };

typedef struct thread_param {
  uint64_t interval;
  int priority;
  int affinity_cpu;
  uint64_t latency_threshold;

  int emit_signal;
  int enable_diff_ts;
  int enable_receive_tracemark;

  uint64_t start_ts;

} thread_param_t;

typedef struct main_params {
  int refresh_rate;
  int verbose;
  int enable_tracing;
  int enable_xdp_tracing;
} main_param_t;

static void process_options(int argc, char *argv[]);
static void print_histograms();
static void sighand(int sig_num);

// Static variables

static uint64_t kernel_latency_hist[MAX_KERNEL_LATENCY];
static uint64_t jitter_hist[MAX_JITTER];
static uint64_t diff_ts_hist[MAX_DIFF_TS];
static main_param_t main_params;
static thread_param_t thread_params;
static ingress_param_t ingress_params;
static egress_param_t egress_params;
static egress_stat_t egress_stats = {.min_kernel_latency = INT_MAX};
static egress_info_t egress_info;
static ingress_stat_t ingress_stats = {.min_kernel_latency = INT_MAX,
                                       .min_interval = INT_MAX};
static int enable_histograms;
static int enable_timestamps;
static enum TSNTask tsn_task;
static struct timespec measures_start;
static struct timespec measures_end;
static struct timespec emit_signal_next;
static pthread_mutex_t emit_signal_mutex;
static pthread_cond_t emit_signal_ts_received;

static int64_t min_diff_ts = INT_MAX;
static int64_t max_diff_ts = 0;
static int64_t avg_diff_ts = 0;
static uint64_t high_diff_ts = 0;

static int debug_latencies = 0;
static int interval_us;

static int gpio86_fd;
static int gpio84_fd;
static int gpio86_state;
static int gpio84_state;
static int received_reverse_motor = 0;

static char ts_tracemark_buf[64];

static void help(char *argv[]) {
  printf(
      "Usage: %s [-f IF -a CPU -p PRIO -i USEC -r USEC] [-b CLIENT_IP] [-d "
      "BUF_LEN -cgtv] [-T LATENCY_THRESHOLD -G] [-xX]\n\n"
      "       -h                 Show help\n"
      "       -f IF              Set the network interface to be used\n"
      "       -a CPU             Pin the real time thread to CPU\n"
      "       -p PRIO            RT thread priority\n"
      "       -i USEC            RT thread wake-up interval (default: 10ms)\n"
      "       -r USEC            non-RT main thread refresh interval\n"
      "       -d BUF_LEN         Set the length of tx buffer\n"
      "       -c                 Receive timestamp and emit signal\n"
      "       -s NS              Common start time reference\n"
      "       -C                 Receive timestamp and print difference with current time\n"
      "       -b CLIENT_IP       Server side RTT\n"
      "       -g                 Print histograms to sdtout on exit\n"
      "       -t                 Enable timestamps\n"
      "       -x                 Use AF_XDP sockets\n"
      "       -D                 Debug latencies\n"
      "       -X                 Trace during XDP packet reception\n"
      "       -T THRESHOLD       Enable tracing until THRESHOLD is reached\n"
      "       -M                 Send tracemark when packet is received\n"
      "       -S                 Send tracemark containing RX SO_TIMESTAMP timestamp\n"
      "       -v                 Verbose\n"
      "\n",
      argv[0]);
}

static void *emit_signal_thread(void *p) {
  (void)p;
  cpu_set_t mask;
  struct timespec current;
  struct timespec previous_emit, previous_ts;
  int64_t emit_diff, ts_diff;
  int latency_spike = 0;
  int ret;
  int reverse_motor = 0;

  // Set thread CPU affinity
  if (thread_params.affinity_cpu) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  pthread_mutex_lock(&emit_signal_mutex);

  for (int i = 0;;i++) {
    pthread_cond_wait(&emit_signal_ts_received, &emit_signal_mutex);

    if(received_reverse_motor) {
      received_reverse_motor = 0;
      reverse_motor = 1;
      continue;
    }

    ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &emit_signal_next, NULL);
    if (ret) {
      fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
      exit(EXIT_FAILURE);
    }

    if(reverse_motor) {
      gpio84_state = toggle_gpio(gpio84_fd, gpio84_state);
      reverse_motor = 0;
    }
    else {
      gpio86_state = toggle_gpio(gpio86_fd, gpio86_state);
      usleep(3);
      gpio86_state = toggle_gpio(gpio86_fd, gpio86_state);
    }

    clock_gettime(CLOCK_REALTIME, &current);
    
    if(debug_latencies) {
      //Check if something went wrong
      if(i > 0) {
        emit_diff = calcdiff_ns_signed(current, previous_emit);
        ts_diff   = calcdiff_ns_signed(emit_signal_next, previous_ts);
        if((emit_diff < ((int64_t)thread_params.interval) - ERROR_MARGIN_NS) ||
           (emit_diff > ((int64_t)thread_params.interval) + ERROR_MARGIN_NS)) {
          fprintf(stderr, "Signal emission interval reached error threshold: %" PRIi64 "\n", emit_diff);
          latency_spike = 1;
        }
        if((ts_diff < ((int64_t)thread_params.interval) - ERROR_MARGIN_NS) ||
           (ts_diff > ((int64_t)thread_params.interval) + ERROR_MARGIN_NS)) {
          fprintf(stderr, "Timestamp interval reached error threshold: %" PRIi64 "\n", ts_diff);
          latency_spike = 1;
        }
        if(latency_spike) {
          fprintf(stderr, "Exiting... Current interval: %d\n", interval_us);
          exit(EXIT_FAILURE);
        }
      }
    }
    previous_emit = current;
    previous_ts = emit_signal_next;
  }

  pthread_mutex_unlock(&emit_signal_mutex);
  pthread_exit(NULL);
}

/* Real-time thread:
 *   - Measures intervals between packet receptions
 */
static void *tsn_thread(void *p) {
  (void)p;
  struct timespec current, previous, next;
  cpu_set_t mask;
  char tracemark_message[128];
  
  // Set thread CPU affinity
  if (thread_params.affinity_cpu) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  clock_gettime(CLOCK_MONOTONIC, &measures_start);

  if (tsn_task == XDP_TASK) setup_poll_fd();

  if (thread_params.start_ts) {
    clock_gettime(CLOCK_REALTIME, &next);
    if (thread_params.start_ts < ts_to_uint(next)) {
      fprintf(stderr, "start timestamp is in the past, aborting...\n");
      exit(EXIT_FAILURE);
    }
    if (thread_params.start_ts > (ts_to_uint(next) + UINT64_C(3600000000000))) {
      fprintf(stderr, "start timestamp is too high, aborting...\n");
      exit(EXIT_FAILURE);
    }
    next = uint_to_ts(thread_params.start_ts);
    add_ns(&next, thread_params.interval);
  }

  if(thread_params.start_ts)
    printf("Will start at %" PRIu64 "\n", ts_to_uint(next));

  // Packet receiving loop
  for (ingress_stats.packets_received = 0;; ingress_stats.packets_received++) {
    // RTT
    if (tsn_task == RTT_TASK) {
      recv_udp_packet();
      send_udp_packet("", 0, &egress_info);

      // Receive packet
    } else if (tsn_task == RECV_PACKET_TASK || tsn_task == XDP_TASK) {

      // Receive UDP or XDP packet
      if (tsn_task == RECV_PACKET_TASK)
        recv_udp_packet();
      else
        recv_xdp_packet(next);

      // Get time for statistics
      clock_gettime(CLOCK_REALTIME, &current);
      recv_xdp_cleanup();
      interval_us = calcdiff_ns(current, previous) / 1000;

      if (thread_params.start_ts)
        add_ns(&next, thread_params.interval);

      if(thread_params.enable_receive_tracemark) {
        sprintf(ts_tracemark_buf, "%" PRIu64, ts_to_uint(current));
        tracemark(ts_tracemark_buf);
      }

      // Get signal timestamp
      if (thread_params.emit_signal) {
        uint64_t emit_signal_t;

        if(tsn_task == XDP_TASK)
          emit_signal_t = decode(ingress_stats.xdp_data);
        else
          emit_signal_t = decode(ingress_stats.data);

        if(emit_signal_t < UINT64_C(1576800000000000000) && emit_signal_t != 17)
          continue;

        pthread_mutex_lock(&emit_signal_mutex);
        if(emit_signal_t == 17)
          received_reverse_motor = 1;
        else
          emit_signal_next = uint_to_ts(emit_signal_t);
        pthread_cond_signal(&emit_signal_ts_received);
        pthread_mutex_unlock(&emit_signal_mutex);
      }

      // Update packet latency estimation statistics
      if(thread_params.enable_diff_ts) {
        uint64_t send_ts;
        if(tsn_task == XDP_TASK)
          send_ts = decode(ingress_stats.xdp_data);
        else
          send_ts = decode(ingress_stats.data);

        int64_t diff_us = (((int64_t) ts_to_uint(current)) - ((int64_t)send_ts)) / 1000;

        min_diff_ts = _min_(diff_us, min_diff_ts);
        max_diff_ts = _max_(diff_us, max_diff_ts);
        avg_diff_ts = (avg_diff_ts * ingress_stats.packets_received + diff_us) / (ingress_stats.packets_received + 1);

        // Histogram
        if (enable_histograms) {
          if (diff_us > ((int)MAX_DIFF_TS) || diff_us < 0)
            high_diff_ts++;
          else
            diff_ts_hist[diff_us]++;
        }

        // If the latency hits the tracing threshold, stop tracing
        if (main_params.enable_tracing &&
            (max_diff_ts > ((int64_t)thread_params.latency_threshold))) {
          sprintf(tracemark_message, "Threshold hit: %" PRIi64 "\n",
                  max_diff_ts);
          tracemark(tracemark_message);
          tracing(0);
          printf(tracemark_message);
          exit(EXIT_SUCCESS);
        }
      }
      else {
        // Update stats
        if (ingress_stats.packets_received) {

          ingress_stats.min_interval =
              _min_(interval_us, ingress_stats.min_interval);
          ingress_stats.max_interval =
              _max_(interval_us, ingress_stats.max_interval);
          ingress_stats.avg_interval =
              (ingress_stats.avg_interval * ingress_stats.packets_received +
               interval_us) /
              (ingress_stats.packets_received + 1);

          // Histogram
          if (enable_histograms) {
            int dist_to_interval = interval_us - (thread_params.interval / 1000);
            dist_to_interval += MAX_JITTER / 2;

            if (dist_to_interval > ((int)MAX_JITTER) || dist_to_interval < 0)
              ingress_stats.high_jitter++;
            else
              jitter_hist[dist_to_interval]++;
          }

          // If the latency hits the tracing threshold, stop tracing
          int jitter = ingress_stats.max_interval - ingress_stats.min_interval;
          if (main_params.enable_tracing &&
              (jitter > ((int)thread_params.latency_threshold))) {
            sprintf(tracemark_message, "Jitter threshold hit: %dus\n",
                    jitter);
            tracemark(tracemark_message);
            tracing(0);
            printf(tracemark_message);
            exit(EXIT_SUCCESS);
          }
        }
      }

      previous = current;
    }
  }

  return NULL;
}

static void create_thread(void *(*thread_function)(void *)) {
  pthread_t thread;
  pthread_attr_t attr;
  struct sched_param param;

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  // Set a specific stack size
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  // Use scheduling parameters of attr
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, thread_function, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create packet receiving thread");
}

/* Main thread:
 *   - Has non-real time priority
 *   - Handles the IO and creates real time threads
 */
int main(int argc, char *argv[]) {
  // Initial values
  ingress_stats.min_interval = INT_MAX;
  ingress_stats.avg_interval = 0;
  ingress_stats.max_interval = 0;
  ingress_stats.min_kernel_latency = INT_MAX;
  ingress_stats.avg_kernel_latency = 0;
  ingress_stats.max_kernel_latency = 0;
  ingress_stats.packets_received = 0;

  // Default configuration values
  thread_params.interval = 100000 * 1000;
  thread_params.priority = 98;
  thread_params.emit_signal = 0;
  thread_params.affinity_cpu = 0;
  thread_params.enable_diff_ts = 0;
  thread_params.enable_receive_tracemark = 0;
  thread_params.start_ts = 0;
  main_params.refresh_rate = 50000;
  main_params.verbose = 0;
  main_params.enable_tracing = 0;
  enable_timestamps = 0;
  enable_histograms = 0;
  tsn_task = RECV_PACKET_TASK;
  ingress_params.tx_buffer_len = 1024;
  ingress_params.enable_ts_tracemark = 0;

  // Process bash options
  process_options(argc, argv);

  set_latency_target();

  ingress_params.use_timestamps = enable_timestamps;
  ingress_params.interval = thread_params.interval;

  egress_info.params = &egress_params;
  egress_info.stats = &egress_stats;

  // Init histograms
  if (enable_histograms) {
    memset(kernel_latency_hist, 0, sizeof(kernel_latency_hist));
    memset(jitter_hist, 0, sizeof(jitter_hist));
    memset(diff_ts_hist, 0, sizeof(diff_ts_hist));
  }

  // Enable ftrace
  if (main_params.enable_tracing || main_params.enable_xdp_tracing) open_fds();

  // Catch breaks with sighand to print the histograms or exit XDP
  init_signals(sighand);

  // Initialize the XDP or UDP packet receiving socket
  if (tsn_task == XDP_TASK)
    init_xdp_recv(&ingress_params, &ingress_stats);
  else
    init_udp_recv(&ingress_params, &ingress_stats, enable_histograms,
                  kernel_latency_hist);

  // Initialize the UDP packet sending socket if RTT is measured
  if (tsn_task == RTT_TASK)
    init_udp_send(&egress_info, 0, 0);

  if (thread_params.emit_signal) {
    pthread_mutex_init(&emit_signal_mutex, NULL);
    pthread_cond_init(&emit_signal_ts_received, NULL);

    gpio86_state = enable_gpio(&gpio86_fd, 86);
    gpio84_state = enable_gpio(&gpio84_fd, 84);
  }

  create_thread(tsn_thread);
  create_thread(emit_signal_thread);

  // Verbose loop
  for (;;) {
    usleep(main_params.refresh_rate);

    // Stats printing
    if (main_params.verbose && ingress_stats.packets_received > 1 &&
        (tsn_task == RECV_PACKET_TASK || tsn_task == XDP_TASK)) {
      int jitter = ingress_stats.max_interval - ingress_stats.min_interval;

      if (thread_params.enable_diff_ts) {
        printf("%9" PRIu64 ": D : %8" PRIi64 " %8" PRIi64 " %17" PRIi64,
               ingress_stats.packets_received, 
               min_diff_ts, 
               avg_diff_ts,
               max_diff_ts);
      } else {
        printf("%9" PRIu64 ": J: %5d, I (10us): %3d %3d %3d [%3d]",
               ingress_stats.packets_received, jitter,
               ingress_stats.min_interval / 10, ingress_stats.avg_interval / 10,
               ingress_stats.max_interval / 10, (int)ingress_stats.high_jitter);
      }

      if (enable_timestamps)
        printf(", K: %4d %4d %4d [%3d]\n", ingress_stats.min_kernel_latency,
               ingress_stats.avg_kernel_latency,
               ingress_stats.max_kernel_latency,
               (int)ingress_stats.high_kernel_latency);
      else
        printf("\n");

      printf("\033[%dA", 1);
    }
  }

  pthread_mutex_destroy(&emit_signal_mutex);
  pthread_cond_destroy(&emit_signal_ts_received);
  close_xdp_socket();
}

/* Print histograms in .json format
 */
static void print_histograms() {
  uint64_t duration;
  int duration_hour, duration_minutes;
  int max_latency, max_jitter, min_jitter;

  clock_gettime(CLOCK_MONOTONIC, &measures_end);

  duration = calcdiff_ns(measures_end, measures_start);
  duration_hour = duration / NSEC_PER_SEC / 3600;
  duration_minutes = duration / NSEC_PER_SEC / 60 - duration_hour * 60;

  if (thread_params.enable_diff_ts) {
    printf("# RX latency histogram\n");

    max_latency = histogram_max(diff_ts_hist, MAX_DIFF_TS - 1);

    for (int j = 0; j <= max_latency; j++)
      printf(" %06d %015" PRIi64 "\n", j, diff_ts_hist[j]);

    printf(
        "# Duration: %dh%d\n"
        "# Lost packets: %" PRIu64 "\n"
        "# Min: %" PRIi64 "\n"
        "# Max: %" PRIi64 "\n"
        "# Avg: %" PRIi64 "\n",
        duration_hour, duration_minutes, high_diff_ts, min_diff_ts, max_diff_ts, avg_diff_ts);
  }
  else {

    max_jitter = histogram_max(jitter_hist, MAX_JITTER - 1);
    min_jitter = histogram_min(jitter_hist, MAX_JITTER - 1);

    if (enable_timestamps) {
      max_latency = histogram_max(kernel_latency_hist, MAX_KERNEL_LATENCY - 1);

      printf("# Packet RX timestamps histogram\n");

      for (int j = 0; j <= max_latency; j++)
        printf(" %06d %015" PRIi64 "\n", j, kernel_latency_hist[j]);

      printf(
          "# Duration: %dh%d\n"
          "# Lost packets: %" PRIu64 "\n",
          duration_hour, duration_minutes, ingress_stats.high_jitter);

    }

    printf("# Packet jitter histogram\n");

    for (int j = min_jitter; j <= max_jitter; j++)
      printf(" %06d %015" PRIi64 "\n", j, jitter_hist[j]);

    printf(
        "# Middle: %d\n"
        "# Duration: %dh%d\n"
        "# Lost packets: %" PRIu64 "\n",
        MAX_JITTER / 2 - min_jitter, duration_hour, duration_minutes,
        ingress_stats.high_jitter);
  }

}

static void sighand(int sig_num) {
  (void)sig_num;

  if (enable_histograms) {
    print_histograms();
  }

  if (tsn_task == XDP_TASK) {
    close_xdp_socket();
  }

  if (ingress_stats.high_jitter)
    fprintf(stderr, "%d packets were lost\n", ((int)ingress_stats.high_jitter));

  exit(EXIT_SUCCESS);
}

/* Process bash options
 */
static void process_options(int argc, char *argv[]) {
  int network_if_specified = 0;

  for (;;) {
    int c = getopt(argc, argv, "a:b:cCs:d:f:ghi:p:r:tvxDXT:GMS");

    if (c == -1) break;

    switch (c) {
      case 'a':
        thread_params.affinity_cpu = atoi(optarg);
        break;
      case 'b':
        tsn_task = RTT_TASK;
        strcpy(egress_params.server_ip, optarg);
        break;
      case 'c':
        thread_params.emit_signal = 1;
        break;
      case 'C':
        thread_params.enable_diff_ts = 1;
        break;
      case 's':
        thread_params.start_ts = strtoull(optarg, NULL, 10);
        break;
      case 'd':
        ingress_params.tx_buffer_len = atoi(optarg);
        if (ingress_params.tx_buffer_len < 1) {
          fprintf(stderr, "BUF_LEN should be greater than 1\n");
          exit(EXIT_FAILURE);
        }
        break;
      case 'f':
        network_if_specified = 1;
        strcpy(ingress_params.network_if, optarg);
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'i':
        thread_params.interval = strtoull(optarg, NULL, 10) * 1000;
        break;
      case 'g':
        enable_histograms = 1;
        break;
      case 'p':
        thread_params.priority = atoi(optarg);
        break;
      case 'r':
        main_params.refresh_rate = atoi(optarg);
        break;
      case 't':
        enable_timestamps = 1;
        break;
      case 'v':
        main_params.verbose = 1;
        break;
      case 'x':
        tsn_task = XDP_TASK;
        break;
      case 'D':
        debug_latencies = 1;
        break;
      case 'X':
        main_params.enable_xdp_tracing = 1;
        break;
      case 'T':
        main_params.enable_tracing = 1;
        thread_params.latency_threshold = atoi(optarg);
        break;
      case 'M':
        thread_params.enable_receive_tracemark = 1;
        break;
      case 'S':
        ingress_params.enable_ts_tracemark = 1;
        break;
    }
  }

  if (!network_if_specified) {
    fprintf(stderr, "You need to specifiy an network interface\n");
    help(argv);
    exit(EXIT_FAILURE);
  }

  if (argc != optind) {
    fprintf(stderr, "Too many arguments\n");
    help(argv);
    exit(EXIT_FAILURE);
  }
}
